import 'dart:async';

import 'package:chopper/chopper.dart';
import 'package:libaqueous/src/api/client.dart';
import 'package:libaqueous/src/api/room.dart';
import 'package:libaqueous/src/api/store.dart';
import 'package:libaqueous/src/api/user.dart';
import 'package:libaqueous/src/client/default_room.dart';
import 'package:libaqueous/src/client/default_user.dart';
import 'package:libaqueous/src/rest/api.dart';
import 'package:libaqueous/src/rest/sync/model/invited_room_sync.dart';
import 'package:libaqueous/src/rest/sync/model/room_sync.dart';
import 'package:libaqueous/src/rest/sync/model/sync_response.dart';
import 'package:rxdart/rxdart.dart';

mixin DefaultClientEmitter {
  final _roomsSubject = BehaviorSubject<List<Room>>();

  Observable<List<Room>> get roomsObservable => _roomsSubject.stream;
}

class DefaultClient with DefaultClientEmitter implements Client {
  DefaultClient(String homeserver, userID, token, this.store)
      : api = MatrixAPI(homeserver, token),
        id = DateTime.now().millisecondsSinceEpoch,
        userID = userID;

  String get homeServer => api.baseUrl;

  final String userID;

  String get token => api.token;

  final int id;

  bool _isRunning = false;

  bool get isRunning => _isRunning;

  int _transactionCount = 0;

  String get transactionID =>
      id.toString() + "q" + (++_transactionCount).toString();

  MatrixAPI api;

  User get myUser => users[userID];

  bool _running = false;

  Store store;

  Map<String, DefaultRoom> _rooms = {};
  Map<String, DefaultUser> _users = {};

  Map<String, Room> get rooms => _rooms;

  Map<String, DefaultUser> get users => _users;

  Future<void> loadFromStore() async {
    final roomDatas = await store.roomDataBulk();

    roomDatas?.forEach((String id, DefaultRoomData roomData) {
      _rooms[id] = DefaultRoom.fromData(this, roomData);
    });

    await getUser(userID);

    // Load users that are in direct chat.
    for (final room in rooms.values) {
      if (room.directUserID != null) {
        // What is this
        await getUser(room.directUserID);
      }
    }

    _roomsSubject.add(rooms.values.toList());
  }

  Future<void> startSync() async {
    if (_running) return;
    _running = true;

    await _sync();
  }

  void stopSync() {
    _running = false;
  }

  Future<void> logout() async {
    await api.loginAPI.postLogout();
  }

  Future<void> _sync() async {
    final stream = _getEventStream();

    await for (SyncResponse data in stream) {
      await _processSyncResponse(data);
    }
  }

  Stream<SyncResponse> _getEventStream() async* {
    while (_running) {
      final syncToken = await store.syncToken() ?? "";
      print("_getEventStream(): sync token: " + syncToken);

      final Response<SyncResponse> resp = await api.syncAPI.sync(
          "{\"room\":{\"state\":{\"lazy_load_members\":true}}}",
          syncToken,
          syncToken.isEmpty,
          30000);
      final syncResp = resp.body;
      syncResp.previousBatch = syncToken;

      yield syncResp;

      await store.setSyncToken(syncResp.nextBatch);
    }
  }

  Future<void> _processSyncResponse(SyncResponse data) async {
    if (data == null) return;

    int changes = ClientChanges.NoChange;

    final isInitialSync = data.previousBatch?.isEmpty ?? true;

    // TODO: Process and store sync response.
    if (data.toDevice?.events?.isNotEmpty ?? false) {
      for (final e in data.toDevice.events) {}
    }

    if (data.accountData?.events?.isNotEmpty ?? false) {
      for (final e in data.accountData.events) {
        switch (e.type) {
          case "m.direct":
            Map<String, DefaultRoom> roomMap = {};
            (Map<String, List>.from(e.content))
                .forEach((String userID, List roomIDs) async {
              for (final roomID in roomIDs) {
                if (!(roomID is String)) break;
                var room = await getRoom(roomID as String);
                if (room == null) {
                  room = DefaultRoom(this, id: roomID);
                  rooms[roomID] = room;
                  changes |= ClientChanges.RoomsChange;
                }
                roomMap[roomID] = room;

                room.d.isDirect = true;
                room.d.directUserID = userID;
              }
            });
            await store.storeRoomDataBulk(
                roomMap.map((id, room) => MapEntry(id, room.d)));
            break;
        }
      }
    }

    if (data.rooms?.join?.isNotEmpty ?? false) {
      changes |= await _processRooms(data.rooms.join, "join",
          isInitialSync: isInitialSync);
    }

    if (data.rooms?.leave?.isNotEmpty ?? false) {
      changes |= await _processRooms(data.rooms.leave, "leave",
          isInitialSync: isInitialSync);
    }

    if (data.rooms?.invite?.isNotEmpty ?? false) {
      await _processInvitedRooms(data.rooms.invite);
    }

    if (changes & ClientChanges.RoomsChange > 0) {
      _roomsSubject.add(rooms.values.toList());
    }
  }

  Future<int> _processRooms(
      Map<String, RoomSync> roomSyncMap, String membership,
      {bool isInitialSync = false}) async {
    int changes = ClientChanges.NoChange;

    Map<String, DefaultRoom> roomMap = {};

    for (final id in roomSyncMap.keys) {
      final roomSync = roomSyncMap[id];

      var room = await getRoom(id);
      if (room == null) {
        room = DefaultRoom(this, id: id);
        rooms[id] = room;
        changes |= ClientChanges.RoomsChange;
      }

      room.processRoomSync(roomSync, isInitialSync: isInitialSync);

      roomMap[id] = room;
    }

    await store.storeRoomDataBulk(roomMap.map((String id, DefaultRoom room) =>
        MapEntry<String, DefaultRoomData>(id, room.d)));

    return changes;
  }

  Future<void> _processInvitedRooms(
      Map<String, InvitedRoomSync> roomSyncMap) async {
    for (final id in roomSyncMap.keys) {
      final roomSync = roomSyncMap[id];

      var room = await getRoom(id);
      if (room == null) {
        room = DefaultRoom(this, id: id);
        rooms[id] = room;
      }

      room.d.userMembership = "invite";
    }
  }

  Future<DefaultRoom> getRoom(String id) async {
    if (_rooms.containsKey(id)) return _rooms[id];

    if (await store.containsRoomData(id)) {
      final room = DefaultRoom.fromData(this, await store.roomData(id));
      _rooms[id] = room;
      return room;
    }

    return null;
  }

  Future<DefaultUser> getUser(String id) async {
    if (_users.containsKey(id)) return _users[id];

    if (await store.containsUserData(id)) {
      final user = DefaultUser.fromData(this, await store.userData(id));
      _users[id] = user;
      return user;
    }

    return null;
  }
}
