import 'package:libaqueous/src/api/client.dart';
import 'package:libaqueous/src/api/data/data_holder.dart';
import 'package:libaqueous/src/api/data/timeline.dart';
import 'package:libaqueous/src/api/room.dart';
import 'package:libaqueous/src/client/default_user.dart';
import 'package:libaqueous/src/rest/events/model/event.dart';
import 'package:libaqueous/src/rest/events/model/event_type.dart';
import 'package:libaqueous/src/rest/events/model/room_avatar_event.dart';
import 'package:libaqueous/src/rest/events/model/room_canonical_alias_event.dart';
import 'package:libaqueous/src/rest/events/model/room_member_event.dart';
import 'package:libaqueous/src/rest/events/model/room_name_event.dart';
import 'package:libaqueous/src/rest/events/model/room_tag.dart';
import 'package:libaqueous/src/rest/events/model/room_topic_event.dart';
import 'package:libaqueous/src/rest/events/model/tokens_chunk_events.dart';
import 'package:libaqueous/src/rest/sync/model/room_sync.dart';
import 'package:libaqueous/src/rest/sync/model/room_sync_account_data.dart';
import 'package:libaqueous/src/rest/sync/model/room_sync_state.dart';
import 'package:libaqueous/src/rest/sync/model/room_sync_timeline.dart';
import 'package:rxdart/rxdart.dart';

abstract class RoomChanges {
  static const NoChange = 0x0;
  static const NameChange = 0x1;
  static const CanonicalAliasChange = 0x2;
  static const TopicChange = 0x4;
  static const UnreadNotifsChange = 0x8;
  static const AvatarChange = 0x10;
  static const JoinStateChange = 0x20;
  static const TagsChange = 0x40;
  static const MembersChange = 0x80;
  static const AccountDataChange = 0x200;
  static const SummaryChange = 0x400;
  static const ReadMarkerChange = 0x800;
  static const OtherChange = 0x8000;
  static const AnyChange = 0xFFFF;
}

class DefaultRoomData {
  String id;
  String name;
  String topic;
  Uri avatarUrl;
  String canonicalAlias;
  bool isDirect = false;
  String directUserID;
  Map<String, double> tags = {};
  bool hasFullMembers = false;
  List<String> memberIDs;
  String userMembership;
  int unreadEventCount = 0;
  int notificationCount = 0;
  int highlightCount = 0;
}

mixin DefaultRoomEmitter {
  final _nameSubject = BehaviorSubject<String>();

  Observable<String> get nameObservable => _nameSubject.stream;

  final _topicSubject = BehaviorSubject<String>();

  Observable<String> get topicObservable => _topicSubject.stream;

  final _avatarUrlSubject = BehaviorSubject<Uri>();

  Observable<Uri> get avatarUrlObservable => _avatarUrlSubject.stream;

  final _canonicalAliasSubject = BehaviorSubject<String>();

  Observable<String> get canonicalAliasObservable =>
      _canonicalAliasSubject.stream;

  final _membersSubject = BehaviorSubject<List<String>>();

  Observable<List<String>> get membersObservable => _membersSubject.stream;

  final _timelineSubject = BehaviorSubject<Timeline>();

  Observable<Timeline> get timelineObservable => _timelineSubject.stream;
}

class DefaultRoom
    with DefaultRoomEmitter
    implements Room, DataHolder<DefaultRoomData> {
  DefaultRoom(this._client, {String id}) : _d = DefaultRoomData()..id = id {}

  DefaultRoom.fromData(this._client, this._d) {}

  final Client _client;

  Client get client => _client;

  final DefaultRoomData _d;

  DefaultRoomData get d => _d;

  String get id => d.id;

  String get name => d.name;

  String get topic => d.topic;

  String get displayName =>
      (name ?? (isDirect ? client.users[directUserID]?.name : null)) ??
      canonicalAlias ??
      id;

  Uri get avatarUrl =>
      d.avatarUrl ?? (isDirect ? client.users[directUserID]?.avatarUrl : null);

  String get canonicalAlias => d.canonicalAlias;

  bool get isDirect => d.isDirect;

  String get directUserID => d.directUserID;

  Map<String, double> get tags => d.tags;

  bool get isFavourite =>
      tags?.containsKey(RoomTag.ROOM_TAG_FAVOURITE) ?? false;

  bool get isLowPriority =>
      tags?.containsKey(RoomTag.ROOM_TAG_LOW_PRIORITY) ?? false;

  String get prevToken => timeline.prevToken;

  bool get hasFullMembers => d.hasFullMembers;

  List<String> get memberIDs => d.memberIDs;

  String get userMembership => d.userMembership;

  int get unreadEventCount => d.unreadEventCount;

  int get notificationCount => d.notificationCount;

  int get highlightCount => d.highlightCount;

  final Timeline _timeline = Timeline();

  Timeline get timeline => _timeline;

  Future<void> processRoomSync(RoomSync roomSync,
      {bool isInitialSync = false}) async {
    int changes = RoomChanges.NoChange;

    int highlightCount = roomSync.unreadNotifications?.highlightCount ?? 0;
    int notificationCount =
        roomSync.unreadNotifications?.notificationCount ?? 0;
    String prevToken = roomSync.timeline?.prevToken ?? "";
    bool limited = roomSync.timeline?.limited ?? false;

    changes |= await processRoomSyncState(roomSync.state);
    changes |= await processRoomSyncTimeline(roomSync.timeline);
    changes |= await processRoomSyncAccountData(roomSync.accountData);

    if (notificationCount != d.notificationCount) {
      d.notificationCount = notificationCount;
    }

    if (highlightCount != d.highlightCount) {
      d.highlightCount = highlightCount;
    }

    if (isInitialSync) {
      await _client.store.storeBackToken(id, prevToken);
    }

    if (changes & RoomChanges.MembersChange > 0) {
      print("Members changed for room $displayName");
    }
    if (changes & RoomChanges.TagsChange > 0) {
      print("Account data changed for room $displayName");
    }

    _timelineSubject.add(timeline);
  }

  Future<int> processRoomSyncState(RoomSyncState syncState) async {
    if (syncState?.events?.isEmpty ?? true) return RoomChanges.NoChange;

    int changes = RoomChanges.NoChange;

    for (final event in syncState.events) {
      changes |= await processStateEvent(event);
    }

    return changes;
  }

  Future<int> processStateEvent(Event event) async {
    if (event is RoomNameEvent) {
      d.name = event.content.name;
      return RoomChanges.NameChange;
    }
    if (event is RoomTopicEvent) {
      d.topic = event.content.topic;
      return RoomChanges.TopicChange;
    }
    if (event is RoomAvatarEvent) {
      d.avatarUrl = event.content.url;
      return RoomChanges.AvatarChange;
    }
    if (event is RoomCanonicalAliasEvent) {
      d.canonicalAlias = event.content.alias;
      return RoomChanges.CanonicalAliasChange;
    }
    if (event is RoomMemberEvent) {
      return await processRoomMemberEvent(event);
    }
    return RoomChanges.NoChange;
  }

  Future<int> processRoomMemberEvent(RoomMemberEvent event) async {
    if (event.stateKey?.isEmpty ?? true) return RoomChanges.NoChange;

    final userID = event.stateKey;

    // TODO: Terrible logic.
    if (!client.users.containsKey(userID) &&
        !await client.store.containsUserData(userID)) {
      final user = DefaultUser.fromData(
          this.client,
          DefaultUserData()
            ..id = userID
            ..avatarUrl = event.content.avatarUrl
            ..name = event.content.displayName);
      client.users[userID] = user;
      await client.store.storeUserData(userID, user.d);
    } else {
      final DefaultUser user = await client.getUser(event.stateKey);

      user.d.avatarUrl = event.content.avatarUrl;
      user.d.name = event.content.displayName;

      await client.store.storeUserData(userID, user.d);
    }
    return RoomChanges.MembersChange;
  }

  Future<int> processRoomSyncTimeline(RoomSyncTimeline syncTimeline) async {
    if (syncTimeline?.events?.isEmpty ?? true) return RoomChanges.NoChange;

    int changes = RoomChanges.NoChange;

    syncTimeline.events.first.mToken =
        syncTimeline.prevToken; // Set mToken for first event.

    if (timeline.prevToken == null) {
      timeline.prevToken = syncTimeline.prevToken;
    }

    for (final event in syncTimeline.events) {
      event.roomID = id; // VERY IMPORTANT

      if (EventType.isStateEvent(event.type)) {
        changes |= await processStateEvent(event);
      }
      await processTimelineEvent(event, false);
    }

    final tokensChunkEvents = TokensChunkEvents()
      ..events = syncTimeline.events
      ..start = syncTimeline.prevToken;

    await client.store.storeRoomEvents(id, tokensChunkEvents, false);

    return changes;
  }

  Future<void> processBackTimelineEvents(
      TokensChunkEvents tokensChunkEvents) async {
    if (tokensChunkEvents?.events?.isEmpty ?? true) return;

    for (final event in tokensChunkEvents.events) {
      await processTimelineEvent(event, true);
    }
  }

  Future<void> processTimelineEvent(Event event, bool isHistory) async {
    // Load sender on demand.
    await client.getUser(event.sender);

    if (isHistory) {
      timeline.events.insert(0, event);
    } else {
      timeline.events.add(event);
    }
  }

  Future<int> processRoomSyncAccountData(
      RoomSyncAccountData syncAccountData) async {
    if (syncAccountData?.events?.isEmpty ?? true) RoomChanges.NoChange;

    int changes = RoomChanges.NoChange;

    for (final event in syncAccountData.events) {
      changes |= await processAccountDataEvent(event);
    }

    return changes;
  }

  Future<int> processAccountDataEvent(Event event) async {
    if (event.type == "m.tag") {
      d.tags?.clear();

      if (event.content.rawContent["tags"] is Map) {
        final tags = (event.content.rawContent["tags"]) as Map<String, dynamic>;
        if (tags.isEmpty) return RoomChanges.NoChange;

        tags.forEach((String tag, dynamic data) {
          if (data is Map<String, dynamic>) {
            double order = null;

            // data["order"] should be double or int, but in the legacy spec it is string.
            if (data["order"] is int) {
              order = data["order"].toDouble();
            } else if (data["order"] is double) {
              order = data["order"];
            } else if (data["order"] is String) {
              order = double.tryParse(data["order"]);
            }

            if (order == null) {
              return;
            }

            d.tags[tag] = order;
          }
        });
      }

      return RoomChanges.TagsChange;
    }

    return RoomChanges.NoChange;
  }

  Future<void> postMessage(Map<String, dynamic> body) async {
    await client.api.roomAPI
        .send(this.id, "m.room.message", client.transactionID, body);
  }

  Future<void> postPlainText(String text) async {
    await postMessage({"msgtype": "m.text", "body": text});
  }

  Future<void> postEmote(String text) async {
    await postMessage({"msgtype": "m.emote", "body": text});
  }

  Future<void> getEarlierMessages(int limit) async {
    final tokensChunkEvents =
        await client.store.getEarlierMessages(id, prevToken, limit);
    if (tokensChunkEvents == null) {
      print("Fetching events from homeserver...");
      await getPreviousEvents(limit);
    } else {
      timeline.prevToken = tokensChunkEvents.end;

      await processBackTimelineEvents(tokensChunkEvents);
    }

    _timelineSubject.add(timeline);
  }

  Future<void> getPreviousEvents(int limit) async {
    final resp = (await client.api.roomAPI
            .getMessages(this.id, prevToken, "", "b", limit, ""))
        .body;
    if (resp?.chunk?.isEmpty ?? true) return;

    timeline.prevToken = resp.end;

    final tokensChunkEvents = TokensChunkEvents()
      ..events = resp.chunk
      ..start = resp.start
      ..end = resp.end;

    tokensChunkEvents.events.last.mToken = tokensChunkEvents.end;

    await client.store.storeBackToken(id, tokensChunkEvents.end);

    await client.store.storeRoomEvents(id, tokensChunkEvents, true);

    await processBackTimelineEvents(tokensChunkEvents);
  }

  Future<void> fetchMembers() async {
    print("fetchMembers(): Fetching a complete list of room members...");

    final resp = (await client.api.roomAPI.getMembers(id)).body;
    if (resp?.chunk?.isEmpty ?? true) return;

    d.hasFullMembers = true;
    d.memberIDs = resp.chunk.map((e) => e.stateKey).toList();

    await client.store.storeRoomData(id, d);

    Map<String, DefaultUser> userMap = {};
    for (final e in resp.chunk) {
      if (client.users.containsKey(e.stateKey) ||
          await client.store.containsUserData(e.stateKey)) continue;

      final user = DefaultUser(client);
      user.d.name = e.content.displayName;
      user.d.avatarUrl = e.content.avatarUrl;

      userMap[e.stateKey] = user;
    }

    await client.store.storeUserDataBulk(userMap.map(
            (String id, DefaultUser user) =>
            MapEntry<String, DefaultUserData>(id, user.d)));

    print("fetchMembers(): Fetched room with " +
        userMap.keys.length.toString() +
        " members.");

    // Loading users on demand.
    for (final event in timeline.events) {
      await client.getUser(event.sender);
    }
  }
}
